@@ -13,10 +13,8 @@ Common Options:
 -p[#], --num-cpus=[#]       Use at most [#] cpus
 -r, --recursive                 continue with the next script, if available
 '''
-# See `Environment.__init__' definition #
-# for valid flags            #
-try: import multiprocessing
-except ImportError: pass
+# See `Environment.__init__' definition for valid flags #
+import multiprocessing
 import getopt
 import sys
 try: import sysconfig
@@ -26,15 +24,36 @@ import platform
 import glob
 import signal
 import subprocess
+import pprint
 import functools
 import itertools
 from functools import partial
 import decorator
 from decorator import decorator
-
+from errno import ENOENT
+import logging
 global PROGRAM_NAME
 PROGRAM_NAME = os.path.basename(sys.argv[0])
 
+# set up the module-level logger
+LOGGER = multiprocessing.log_to_stderr()
+LOGGER.setLevel(logging.CRITICAL)
+
+debug = LOGGER.debug
+info = LOGGER.info
+warning = LOGGER.warning
+error = LOGGER.error
+critical = LOGGER.critical
+log = LOGGER.log
+exception = LOGGER.exception
+
+def pformat_list(L):
+    """
+    Takes a list and turns each item into a str
+    then returns the pretty-printed version of that list
+    """
+    return pprint.pformat([str(itm) for itm in L])
+
 class InvalidFileException(ValueError):
     '''
     Exception for files that do not return a valid FilenameParser object
@@ -51,10 +70,8 @@ def exit_on_Usage(func, *args, **kargs):
     try:
         return func(*args, **kargs)
     except Usage, err:
-        print_debug('{0!s}: {1!s}'.format(PROGRAM_NAME, err.msg))
-        if locals().has_key('options'):
-            if options.has_key('h') or options.has_key('help'): return 2
-        print_debug("for help use --help")
+        sys.stderr.write('%s: %s\n' % (PROGRAM_NAME, err.msg))
+        sys.stderr.write("for help use --help\n")
         sys.exit(2)
 
 class Environment(object):
@@ -63,7 +80,6 @@ class Environment(object):
     
     provides an execution environment for jobs
     '''
-    
     def __init__(self, short_opts='', long_opts=[], doc='',
                  version=''):
         self._script_version = version
@@ -85,7 +101,7 @@ class Environment(object):
         self.long_opts.extend(self.verbosity_levels)
         self.long_opts.extend(long_opts)
         self._options = self.parse_argv()
-        self._set_verbosity()
+        self._set_logging_level()
         self._exit_if_needed()
         
         self._set_initial_num_cpus()
@@ -98,10 +114,10 @@ class Environment(object):
         '''
         if options.has_key('target'):
             if len(options['target'].strip()) is not 0:
-                if self._target_dir is not None and verbose:
-                    print_debug('Overriding target dir', self._target_dir,
-                                'with user-specified output directory',
-                                options['target'])
+                if self._target_dir is not None:
+                    info('Overriding target dir %s '
+                         'with user-specified output directory %s',
+                          self._target_dir, options['target'])
                 self._target_dir = options['target']
 
     def get_target_dir(self):
@@ -109,8 +125,7 @@ class Environment(object):
 
     def set_target_dir(self, target_dir):
         self.target_dir = target_dir
-        if self.is_debug():
-            print_debug('Setting Environment target_dir to {0!s}'.format(target_dir))
+        debug('Setting Environment target_dir to %s', target_dir)
         return
         
     def get_source_dir(self):
@@ -118,8 +133,7 @@ class Environment(object):
 
     def set_source_dir(self, source_dir):
         self.source_dir = source_dir
-        if self.is_debug():
-            print_debug('Setting Environment source_dir to {0!s}'.format(source_dir))
+        debug('Setting Environment source_dir to %s', source_dir)
         return
 
     @exit_on_Usage
@@ -134,10 +148,9 @@ class Environment(object):
         if source_dir is None:
             raise Usage('Cannot use --find without specifying a source',
                         'directory')
-        if self.is_debug():
-            print_debug('Searching for valid files')
+            debug('Searching for valid files')
             if self.allowed_extensions is not None:
-                print_debug('Valid file extensions are',
+                debug('Valid file extensions are %s',
                             ' '.join(self.allowed_extensions))
         filename_parser = self.get_filename_parser()
         parsed_filenames = []
@@ -174,16 +187,23 @@ class Environment(object):
         updates _sequence with additional_sequence and
         _unprocessed_sequence (if applicable)
         '''
-        if self.is_debug():
-            print_debug('Updating sequence...')
-        if self.is_debug():
-            print_debug('Checking for user-specified files')
-            if self.allowed_extensions is not None:
-                print_debug('Valid file extensions are',
-                            ' '.join(self.allowed_extensions))
+        debug('Updating sequence of files...')
+        debug('Checking for user-specified files...')
+        # process wildcards as needed
+        unprocessed_files = []
+        for item in _iter_except(self._unprocessed_sequence.pop, IndexError):
+            files = glob.glob(item)
+            if len(files) == 1: unprocessed_files.append(files[0])
+            elif len(files) > 1: unprocessed_files.extend(files)
+        if self.allowed_extensions is not None:
+            debug('Valid file extensions are %s',
+                  ' '.join(self.allowed_extensions))
+        if len(unprocessed_files) > 0:
+            debug('Found the following files:')
+            debug(pformat_list(unprocessed_files))
         filename_parser = self.get_filename_parser()
         # note, filenames get processed backward
-        files = itertools.chain(_iter_except(self._unprocessed_sequence.pop,
+        files = itertools.chain(_iter_except(unprocessed_files.pop,
                                              IndexError),
                                 _iter_except(additional_sequence.pop,
                                              IndexError))
@@ -208,11 +228,10 @@ class Environment(object):
         options = self._options
         # check for help
         if options.has_key('h') or options.has_key('help'):
-            raise Usage(os.linesep.join([PROGRAM_NAME, __doc__,
-                                         self._script_doc]))
+            raise Usage('%s\n%s' % (__doc__, self._script_doc))
         # check for version info
         elif options.has_key('v') or options.has_key('version'):
-            raise Usage(PROGRAM_NAME, ': version ', self._script_version)
+            raise Usage('version %s' % self._script_version)
         return
     
     def _set_initial_num_cpus(self):
@@ -373,27 +392,19 @@ class Environment(object):
         return options
 
     @exit_on_Usage
-    def _set_verbosity(self):
+    def _set_logging_level(self):
         options = self._options
         verbosity_levels = self.verbosity_levels
-        # check verbosity
-        for verbosity in verbosity_levels:
-            setattr(self, '_' + verbosity, options.has_key(verbosity))
-
-        if sum([getattr(self, '_' + verbosity) for
-                verbosity in verbosity_levels]) > 1:
-            raise Usage('can only specify at most one of ', 
-                        ', '.join(['--' + x for x in verbosity_levels]))
-        # debug implies verbose
-        if self._debug: self._verbose = True
+        if options.has_key('debug'): self._logging_level = logging.DEBUG
+        elif options.has_key('verbose'): self._logging_level = logging.INFO
+        elif options.has_key('quiet'): self._logging_level = logging.WARNING
+        elif options.has_key('silent'): self._logging_level = logging.ERROR
+        else: self._logging_level = logging.WARNING # default
+        LOGGER.setLevel(self.logging_level())
         return
 
     def get_verbose_kwargs(self):
-        verbose_kwargs = {'silent': self.is_silent(),
-                          'quiet': self.is_quiet(),
-                          'verbose': self.is_verbose(),
-                          'debug': self.is_debug()}
-        return verbose_kwargs
+        return {'logging_level': self.logging_level()}
 
     def disable_action(self):
         '''
@@ -423,17 +434,14 @@ class Environment(object):
         if with_args is True, then partial is used to apply arguments as
         appropriate
         '''
-        
         if with_args:
-            if self.is_debug():
-                print_debug('Using {0!s} with kwargs {1!s}'.format(
-                                self._filename_parser.__name__,
-                                self.get_fp_kwargs()))
+            debug('Using %s with the following kwargs:',
+                  self._filename_parser.__name__)
+            debug('\n%s', pprint.pformat(self.get_fp_kwargs()))
             return partial(self._filename_parser, **self.get_fp_kwargs())
         else:
-            if self.is_debug():
-                print_debug('Using {0!s} without kwargs'.format(
-                                            self._filename_parser.__name__))
+            debug('Using %s without kwargs',
+                  self._filename_parser.__name__)
             return self._filename_parser
 
     def set_num_cpus(self, num):
@@ -451,18 +459,6 @@ class Environment(object):
         '''
         return self._num_cpus
 
-    def is_silent(self):
-        return self._silent
-    
-    def is_quiet(self):
-        return self._verbose
-
-    def is_verbose(self):
-        return self._verbose
-
-    def is_debug(self):
-        return self._debug
-
     def get_action_kwargs(self):
         '''
         get the kwargs (a dictionary) for action
@@ -505,13 +501,14 @@ class Environment(object):
         return dict((item for item in kwargs_items))
    
     def _is_valid_file(self, f):
-        '''checks if a file is valid for processing'''
+        '''
+        checks if a file is valid for processing
+        '''
         if not os.path.isfile(f):
-            if self.is_debug():
-                print_debug('Skipping', _quote(f), '. It is not a file.')
+            debug('Skipping %s. It is not a file.', _quote(f))
             return False
         elif f.startswith('.'):
-            print_debug('Skipping hidden file', _quote(f))
+            debug('Skipping hidden file %s', _quote(f))
             return False
         elif self.allowed_extensions is None:
             return True
@@ -520,11 +517,13 @@ class Environment(object):
             if file_ext in self.allowed_extensions:
                 return True
             else:
-                if self.is_debug():
-                    print_debug('Skipping', _quote(f), 'because file',
-                                'does not have a valid file extension')
+                debug('Skipping %s because file does not have a valid file '
+                      'extension', _quote(f))
                 return False
     
+    def logging_level(self):
+        return self._logging_level
+    
     @exit_on_Usage 
     def do_action(self, action, stay_open=False):
         '''
@@ -533,18 +532,19 @@ class Environment(object):
         actions should be functions that at least take FilenameParser objects
         '''
         sequence = list(self.get_sequence())
+        
         if len(sequence) == 0:
             raise Usage('No input files specified or found. Nothing to do.')
         if not self._allow_action:
-            print_debug('Test run. Nothing done.')
-            print_debug('I would have acted on these files:',
-                        ', '.join([str(f) for f in sequence]))
-            return self.execute_next_script()
+            info('Test run. Nothing done.')
+            info('I would have acted on the following files:') 
+            info(pformat_list(sequence))
+            sys.exit(0)
         
         kwargs = self.get_action_kwargs()
 
         max_cpus = len(sequence)
-        if self.is_debug(): print_debug('Debugging mode enabled')
+        debug('Debugging mode enabled')
     
         used_cpus = min([self.get_num_cpus(), max_cpus])
     
@@ -553,37 +553,35 @@ class Environment(object):
             item.check_output_dir(item.output_dir)
         
         if used_cpus == 1:
-            if self.is_debug(): print_debug('multiprocessing disabled')
+            debug('multiprocessing disabled')
             for item in sequence:
                 stdout = action(item, **kwargs)
-                if not self.is_silent() and stdout is not None:
+                if not LOGGER.getEffectiveLevel() >= 50 and stdout is not None:
                     print >>sys.stdout, stdout
         else:
-            if self.is_debug():
-                print_debug('WARNING: multiprocessing enabled', os.linesep,
-                            'debug output may get mangeled')
-            if self.is_verbose():
-                print_debug('Using', str(used_cpus), 'cpu(s)...')
             if not platform.system() == 'Windows':
                 signal.signal(signal.SIGCHLD, signal.SIG_DFL)
+            debug('multiprocessing enabled')
             p = multiprocessing.Pool(processes=used_cpus)
+            debug('Initialized pool of %d workers', used_cpus)
             results = [p.apply_async(action, (item,), kwargs) for
                        item in sequence]
             stdouts = (result.get() for result in results)
             stdouts_good = filter(lambda x: type(x) is str, stdouts)
-            if not self.is_silent() and stdouts_good is not None:
+            if not self.logging_level() == logging.ERROR and \
+                    stdouts_good is not None:
                 print >>sys.stdout, os.linesep.join(stdouts_good)
 
+        if self.next_script is not None:
+            return self.execute_next_script()
         if not stay_open:
             sys.exit(0)
-        elif self.next_script is not None:
-            return self.execute_next_script()
     
     def execute_next_script(self):
         '''
         execute the next script
         '''
-        if debug: print_debug('Launching the next script', self.next_script)
+        debug('Launching the next script: %s' % self.next_script)
         # always proceed with --find
         next_script_args = ["--find"]
         # pass along verbosity level
@@ -601,41 +599,35 @@ class Usage(Exception):
     def __str__(self):
         return self.msg
 
-def print_debug(*args):
-    statement = ' '.join(map(str, args))
-    print >>sys.stderr, statement
-
 def assert_path(path):
     '''if path does not exist, raise IOError'''
     if path is None: raise IOError('NoneType is not a valid path')
     if os.path.exists(path): return True
     else:
-        raise IOError(errno.ENOENT, os.strerror(errno.ENOENT), path)
+        raise IOError(ENOENT, os.strerror(ENOENT), path)
 
 class FilenameParser(object):
     @exit_on_Usage
     def __init__(self, filename,
                  source_dir=None, target_dir=None,
-                 verbose=False, debug=False,
                  *args, **kwargs):
 
-        self._debug = debug
         self.additional_args = args
         self.__dict__.update(kwargs)
 
-        if debug: print_debug('Parsing filename', filename)
+        debug('Parsing filename %s', filename)
 
         self.set_input_file(filename)
 
         self.input_dir = os.path.split(self.input_file)[0]
         if self.input_dir == '': self.input_dir = os.curdir
-        if debug: print_debug('Using', self.input_dir, 'as input_dir')
+        debug('Using %s as input_dir', self.input_dir)
         self.file_extension = os.path.splitext(
                                 self.input_file)[1].lstrip(os.extsep)
 
         if target_dir is not None: self.target_dir = target_dir
         else: raise Usage('Must specify an output directory with --target')
-        if debug: print_debug('Using', self.target_dir, 'as target_dir')
+        debug('Using %s as target_dir', self.target_dir)
 
         # Make sure we have the right output dir
         # Try to build target_dir/x/y/z from source_dir/x/y/z in input path
@@ -648,7 +640,7 @@ class FilenameParser(object):
         except ValueError:
                 self.output_dir = self.target_dir
        
-        if debug: print_debug('Using', self.output_dir, 'as output_dir')
+        debug('Using %s as output_dir', self.output_dir)
 
         self.protoname = os.path.splitext(
                             os.path.basename(self.input_file))[0]
@@ -660,18 +652,16 @@ class FilenameParser(object):
         return self.input_file
 
     def set_input_file(self, filename):
-        if self._debug: print_debug('Checking for', filename, '...')
+        debug('Checking for %s ...', filename)
         assert_path(filename)
         self.input_file = filename
 
     def check_output_dir(self, output_dir):
         # Make the output directory, complain if we fail
         if os.path.exists(output_dir):
-            if self._debug: print_debug('Output directory', 
-                                    _quote(output_dir), 'already exists')
+            debug('Output directory %s already exists', _quote(output_dir))
         else:
-            if self._debug: print_debug('Creating directory',
-                                        _quote(output_dir))
+            debug('Creating directory %s', _quote(output_dir))
             os.makedirs(output_dir, mode=0755)
             if not os.path.exists(output_dir):
                 raise IOError('Could not create directory ' + output_dir)
@@ -699,7 +689,7 @@ def leaves(dir_or_file, allow_symlinks = True, ignore_hidden_files = True,
 
     # Base case
     if os.path.isfile(dir_or_file) and not dir_or_file.startswith('.'):
-        if debug: print_debug('Found file', dir_or_file)
+        debug('Found file %s', dir_or_file)
         return dir_or_file
 
     # Recurse
@@ -884,3 +874,38 @@ def _iter_except(func, exception, first=None):
             yield func()
     except exception:
         pass
+
+def get_logger(level=logging.WARNING):
+    """
+    get_logger(level=WARNING) wraps multiprocessing.get_logger()
+    
+    adds an AnnounceExitFilter to prevent output from getting very garbled
+    at program exit
+    """
+    logger = multiprocessing.get_logger()
+    logger.addFilter(AnnounceExitFilter(False))
+    logger.setLevel(level)
+    return logger
+
+class AnnounceExitFilter(object):
+    """
+    rejects messages announcing thread exit iff the initial condition is False
+
+    looks for specific messages hardcoded into multiprocessing/pool.py
+    see source for more details
+    """
+    def __init__(self, announce_exit):
+        self._announce_exit = announce_exit
+        super(AnnounceExitFilter, self).__init__()
+
+    def filter(self, record):
+        if record.getMessage().startswith('worker got sentinel') or \
+           record.getMessage().startswith('worker exiting') or \
+           record.getMessage().startswith('recreated blocker') or \
+           record.getMessage().startswith('process shutting down') or \
+           record.getMessage().startswith('running all "atexit"') or \
+           record.getMessage().endswith('closing conn') or \
+           record.getMessage().startswith('DECREF'):
+            return self._announce_exit
+        else:
+            return True
\ No newline at end of file
